######################################################################
# Project: Estimation and Inference for Generalized Geoadditive Models
# Author: Shan Yu, Guannan Wang, Li Wang, Chenhui Liu and Lijian Yang
# Date: November 2018
######################################################################
# Function: gam_bps_poi
# Stage 1 estimation
# Input:
# (1) y: the response variable
# (2) X: the covariate matrix
# (3) BQ2: the bivariate matrix after QR decomposition
# (4) P: penalty matrix
# (5) lambda: the tuning parameter
# (6) off: the offset
# (7) theta: used in th negative binomial
# Output:
# (1) beta_hat: coefficients estimated for univariate functions
# (2) theta_hat: coefficients estimated for bivariate function
# (3) lambdac: selected tuning parameter
# (4) gcv: minimum of GCV
# (5) df: degree of freedom of model
gam_bps=function(y,X=NULL,BQ2,P,lambda,off=0,theta=0){
  nobs=length(y)
  weights=rep(1,nobs)
  nl=length(lambda)
  nobs=length(y)
  if(!is.null(X)){
    d1=dim(X)[2]
    d2=dim(BQ2)[2]
    zero1=matrix(rep(0,d1*d2),ncol=d2)
    zero2=matrix(rep(0,d1*(d1+d2)),ncol=d1)
    D=rbind(zero1,P)
    D=cbind(zero2,D)
    Z=cbind(X,BQ2)
    alpha_all=matrix(rep(0,(d1+d2)*nl),ncol=nl)
  } else{
    d1=0
    d2=dim(BQ2)[2]
    D=P
    Z=BQ2
    alpha_all=matrix(rep(0,d2*nl),ncol=nl)
  }

  gcv_all=rep(0,nl)
  for(il in 1:nl){
    mustart=initial(y)
    etastart=linkfun(mustart)
    if(theta!=0) var=variance(mustart,theta) else
      var=variance(mustart)
    mevg=mu.eta(etastart)
    Y_iter=(etastart-off)+(y-mustart)/mevg
    W_iter=(mevg^2)/var
    W_iter=as.vector(W_iter)
    temp1=as.matrix(W_iter*Z)
    temp2=W_iter*Y_iter
    temp3=crossprod(Z,temp1)+lambda[il]*D
    alpha_old=solve(temp3,crossprod(Z,temp2))
    #print(alpha_old[1:3])
    step=0
    delta1=1
    delta2=1
    while(delta1>1e-5 & sum(is.infinite(delta2))==0 & step<=10 ){
      step=step+1
      eta=Z%*%alpha_old+off
      mu=linkinv(eta)
      if(theta!=0) var=variance(mu,theta) else
        var=variance(mu)
      mevg=mu.eta(eta)
      Y_iter=(eta-off)+(y-mu)/mevg
      W_iter=as.vector((mevg^2)/var)
      temp1=W_iter*Z
      temp2=W_iter*Y_iter
      temp3=crossprod(Z,temp1)+lambda[il]*D
      alpha_new=solve(temp3,crossprod(Z,temp2))
      #print(alpha_new[1:3])
      eta_new=Z%*%alpha_new
      delta1=sqrt(mean((alpha_new-alpha_old)^2))
      delta2=exp(eta_new)
      if(sum(is.infinite(delta2))==0){
        alpha_old=alpha_new
      }
    }
    alpha_all[,il]=as.matrix(alpha_old)

    #calculate gcv
    DD=solve(temp3)
    Slambda=Z%*%tcrossprod(DD,temp1)
    df=sum(diag(Slambda))
    yhat=Z%*%alpha_old
    gcv=nobs*sum(W_iter*(Y_iter-yhat)^2)/(nobs-df)^2;

    gcv_all[il]=gcv
  }
  #plot(log(lambda),gcv_all,type='l')
  j=which.min(gcv_all)
  gcv=gcv_all[j]
  lambdac=lambda[j]
  alpha_hat=alpha_all[,j]
  beta_hat=NULL
  if(d1!=0) beta_hat=alpha_hat[1:d1]
  theta_hat=alpha_hat[(d1+1):(d1+d2)]
  list(beta_hat=beta_hat,theta_hat=theta_hat,
       lambdac=lambdac,gcv=gcv,df=df)
}
