######################################################################
# Project: Estimation and Inference for Generalized Geoadditive Models
# Author: Shan Yu, Guannan Wang, Li Wang, Chenhui Liu and Lijian Yang
# Date: November 2018
######################################################################
# Function: gam_bps
# Stage 1 estimation for negative binomial
# Input:
# (1) y: the response variable
# (2) X: the covariate matrix
# (3) BQ2: the bivariate matrix after QR decomposition
# (4) P: penalty matrix
# (5) lambda: the tuning parameter
# (6) off: the offset
# (7) theta: used in th negative binomial
# Output:
# (1) beta_hat: coefficients estimated for univariate functions
# (2) theta_hat: coefficients estimated for bivariate function
# (3) lambdac: selected tuning parameter
# (4) gcv: minimum of GCV
# (5) df: degree of freedom of model
# (6) theta: estimated theta in the negative binomial
gam_bps_nb=function(y,XX,BQ2,P,lambda,off=0,r.theta){

  lower=r.theta[1]
  upper=r.theta[2]
  result1=gam_bps(y,XX,BQ2,P,lambda,off,upper)
  y_hat=linkinv(XX%*%result1$beta_hat+BQ2%*%result1$theta_hat)
  tt=(y-y_hat)^2/variance(y_hat,upper)
  U_sigma=1/(length(y)-result1$df)*sum(tt)

  result2=gam_bps(y,XX,BQ2,P,lambda,off,lower)
  y_hat=linkinv(XX%*%result2$beta_hat+BQ2%*%result2$theta_hat)
  tt=(y-y_hat)^2/variance(y_hat,lower)
  L_sigma=1/(length(y)-result2$df)*sum(tt)

  M_sigma=2
  step=1
  if ( (L_sigma-1)>0 & (U_sigma-1) >0 ) {
    M_sigma=1
    result=result2
    middle=lower
  }
  if ( (L_sigma-1)<0 & (U_sigma-1) < 0 ) {
    M_sigma=1
    result=result1
    middle=upper
  }

  while(abs(M_sigma-1)>0.01 & step <= 10){
    middle=(upper+lower)/2
    result=gam_bps(y,XX,BQ2,P,lambda,off=0,middle)
    y_hat=linkinv(XX%*%result$beta_hat+BQ2%*%result$theta_hat)
    tt=(y-y_hat)^2/variance(y_hat,middle)
    M_sigma=1/(length(y)-result$df)*sum(tt)
    if((M_sigma-1)*(U_sigma-1)<0){
      lower=middle
      L_sigma=M_sigma
    } else {
      upper=middle
      U_sigma=M_sigma
    }
    print(c(middle,M_sigma))
    step=step+1
  }
  list(beta_hat=result$beta_hat,theta_hat=result$theta_hat,
       lambdac=result$lambdac,gcv=result$gcv,df=result$df,theta=middle)
}
