######################################################################
# Project: Estimation and Inference for Generalized Geoadditive Models
# Author: Shan Yu, Guannan Wang, Li Wang, Chenhui Liu and Lijian Yang
# Date: November 2018
#######################################################################
# Function: ggam: fit the ggam
# input:
# (1) Y: the response variable.
# (2) X: the covariate matrix. Each column represents one covariate.
# (3) S: the location matrix. Each row indicates the loacation of one point.
# (4) Tr: the triangulation matrix. Each row indicates one triangle.
# (5) V: the vertices matrix of the triangulation.
# (6) d: the degree of the bivariate spline
# (7) r: the smoothness of the bivariate spline
# (8) rho: the degree of univariate spline
# (9) knots: the knots matrix of the univariate spline. Usually, it is a N+2 by p
# matrix. The r-th column is the knots of r-th covariate. The first element
# and the last element are the boundary knots.
# (10) N: the number of interior knots
# (11) backfitting: indicator of backfitting. If backfitting=TRUE, we are going to
# use the backfitting procedure.
# (12) lambda: the vector of tuning parameter. We use GCV to select them.
# (13) off: offset value
# (14) family: the exponential famliy. For example, nb_bps() for negative binomial.
# (15) bandwidth.option: the method we use to estimate the bandwidth.
# (16) nb: indicator for negative binomial
# (17) r.theta: the range of theta when we search for theta in the nb.
# Output:
# (1) beta_hat: coefficients estimated for univariate functions
# (2) theta_hat: coefficients estimated for bivariate function
# (3) lambdac: selected tuning parameter
# (4) gcv: minimum of GCV
# (5) df: degree of freedom of model
# (6) rho, d, r, knots, family, bandwidth.option are the same as the arguments
# in the input
# (7) sigma_2: estimated sigma_2
# (8) theta: estimated theta in the negative binomial
# (9) mhat: estimated component functions in the evaluated in the sample points
# (10) y, X: re-order y and X
# (11) H_opt: optimal bandwidth for each component functions

source('gam_bps_nb.R')
source('gam_bps.R')
source('setup_bps.R')
ggam=function(Y,X,S,Tr,V,d=2,r=1,rho=2,knots=NULL,N=4,backfitting=TRUE,
              lambda,off=0,fam.type,bandwidth.option=1,r.theta=c(1,8)){

  family=setup_bps(fam.type)
  variance <<- family$variance
  linkinv <<- family$linkinv
  linkfun <<- family$linkfun
  mu.eta <<- family$mu.eta
  gprime <<- family$gprime
  initial <<- family$initialize

  # generate basis function for S using bivariate spline
  B0=basis(V,Tr,d,r,S)
  H=B0$H # smoothness and qr decomposition
  Q2=qrH(H)
  P=B0$K # energy matrix P
  P=t(Q2)%*%P%*%Q2
  Bi=B0$Bi # bivariate spline basis matrix
  BQ2=as.matrix(Bi%*%Q2)
  Ind=B0$Ind
  y=Y[Ind]
  X=X[Ind,]

  # generate basis function for X using univariate spline
  X.p=dim(X)[2]
  XX=NULL
  p=NULL
  if(is.null(knots)){
    for(alpha in 1:X.p){
      UniBasis=Basis_generator(X[,alpha],N=N,q=rho+1)
      Xalpha=UniBasis$Bx
      knots=cbind(knots,UniBasis$knots)
      XX=cbind(XX,Xalpha)
      p=c(p,dim(Xalpha)[2])
    }
  } else {
    for(alpha in 1:X.p){
      Xalpha=Basis_generator(X[,alpha],N=N,q=rho+1,knots=knots[,alpha])$Bx
      XX=cbind(XX,Xalpha)
      p=c(p,dim(Xalpha)[2])
    }
  }
  XX=as.matrix(XX)

  # Stage 1
  ## fit the model
  if(fam.type=='poi') {
    result=gam_bps(y,XX,BQ2,P,lambda,off)
  } else if(fam.type=='nb'){
    result=gam_bps_nb(y,XX,BQ2,P,lambda,off,r.theta)
  }

  result$rho=rho
  result$knots=knots
  result$d=d
  result$r=r
  result$family=family

  y_hat=linkinv(XX%*%result$beta_hat+BQ2%*%result$theta_hat)
  if(fam.type=='nb'){
    tt=(y-y_hat)^2/variance(y_hat,result$theta)
  }else if(fam.type=='poi'){
    tt=(y-y_hat)^2/variance(y_hat)
    result$theta=0
  }
  sigma_2=1/(length(y)-result$df)*sum(tt)
  result$sigma_2=sigma_2

  ## spline estimators
  beta=result$beta_hat
  p=c(0,cumsum(p))
  mhat=NULL
  for(alpha in 1:X.p){
    m=XX[,(p[alpha]+1):(p[alpha+1])]%*%beta[(p[alpha]+1):(p[alpha+1])]
    mhat=cbind(mhat,m)
  }
  mm=BQ2%*%result$theta_hat
  mhat=cbind(mhat,mm)

  result$y=y
  result$mhat=mhat
  result$X=X

  # Stage 2
  H_opt=NULL
  if(backfitting==TRUE){
    n=length(y)
    for(alpha in 1:X.p){

      ## select optimal bandwidth
      if (fam.type=='nb'){
        a=result$theta
      }else if(fam.type=='poi'){
        a=0
      }
      if(bandwidth.option==1) {
        fun1=key_functionsSBLL(y,X,alpha,mhat,sigma_2,X[,alpha],theta=a)
      } else if (bandwidth.option==2) {
        fun1=key_functionsSBLL2(y,X,alpha,mhat,sigma_2,X[,alpha],theta=a)
      } else {
        fun1=key_functionsSBLL3(y,X,alpha,mhat,sigma_2,X[,alpha],theta=a)
      }

      h_opt=length(y)^(-0.2)*(5*7*sum(fun1$esigma/fun1$df_fixed)/sum(fun1$beta_dev2^2))^(0.2)
      H_opt=c(H_opt,h_opt)
  }
    result$opt.bandwidth=H_opt
    result$bandwidth.option=bandwidth.option
  }
  result
}
