######################################################################
# Project: Estimation and Inference for Generalized Geoadditive Models
# Author: Shan Yu, Guannan Wang, Li Wang, Chenhui Liu and Lijian Yang
# Date: November 2018
######################################################################
rm(list = ls())
# library
library(bindata)
library(mgcv)
library(parallel)
library(splines)
install.packages("http://lilywang.public.iastate.edu/research/Basis_v1.0.tar.gz",repos=NULL,type="source")
library(Basis)

# source files
source('datagenerate_nb.R')
source('datagenerate_poi.R')
source('ggam.R')
source('ggam_pred.R')
source('SBK_locp.R')
source('spline_basis.R')
source('truncated_spline.R')
source('plot_SCB.R')

# beta_1
beta.1=function(x0) 1/2*sin(2*pi*x0)-cos(pi*x0)
# beta_2
beta.2=function(x0) 4*((x0-0.5)^2-2/3*(0.5)^3)
# beta_3
beta.3=function(x0) x0-0.5

# true beta at gridded points
X0=matrix(rep(seq(0.01,0.99,0.01),each=3),ncol=3,byrow=TRUE)
beta0.1=beta.1(X0[,1])
beta0.2=beta.2(X0[,2])
beta0.3=beta.3(X0[,3])

# true alpha at gridded points
fsb=list(fs.boundary())[[1]]
xx=seq(-1,4,0.05)
yy=seq(-1,1,0.05)
S0=expand.grid(xx,yy)
v=S0[,1]
w=S0[,2]
m=fs.test(v,w,b=1)+sin(5*v)+cos(5*w)
names(fsb)=c("v","w")
ind=inSide(fsb,x=v,y=w)
S0=S0[ind,]
alpha0=0.5*m[ind]-0.5

# triangulation information
Tr=as.matrix(read.table('data/Tr1.txt',header=FALSE))
V=as.matrix(read.table('data/V1.txt',header=FALSE))

# penalty parameter
lambda_start=0.001
lambda_end=32
nlambda=10
lambda=exp(seq(log(lambda_start),log(lambda_end),length.out=nlambda))

fam.type='nb'
fam.type='poi'

Simulation<-function(iter){
  #iter=1
  set.seed(iter)
  
  if(fam.type=='nb'){
    Data=DataGenerator_nb(2000)
  }else if(fam.type=='poi'){
    Data=DataGenerator_poi(2000)
  }
  
  fit=ggam(Y=Data$Y,X=Data$X,S=Data$S,Tr=Tr,V=V,N=4,backfitting=TRUE,
           lambda=lambda,off=0,fam.type=fam.type)
  
  pred=ggam.pred(fit,X0,S0,V,Tr,testlevel=0.05)
  
  # the MISE
  aa=c(sum((beta0.1-pred$beta_pred[,1])^2)*0.01,
       sum((beta0.2-pred$beta_pred[,2])^2)*0.01,
       sum((beta0.3-pred$beta_pred[,3])^2)*0.01,
       sum((alpha0[pred$alpha.ind]-pred$alpha_pred)^2)*0.05^2)
  
  # coverage
  ind=seq(10,90,4)
  cover=c(prod((beta0.1>=pred$Lower[,1]& beta0.1<= pred$Upper[,1])[ind])==1,
          prod((beta0.2>=pred$Lower[,2]& beta0.2<= pred$Upper[,2])[ind])==1,
          prod((beta0.3>=pred$Lower[,3]& beta0.3<= pred$Upper[,3])[ind])==1)
  
  # results
  res=c(aa,cover)
  names(res)=c('beta1.MISE','beta2.MISE','beta3.MISE','alpha.MISE','beta1.cover',
               'beta2.cover','beta3.cover')
  
  # SCBs plots
  # beta1
  plot.SCB(X0[,1],pred$beta_pred[,1],pred$Upper[,1],pred$Lower[,1],beta0.1)
  # beta2
  plot.SCB(X0[,2],pred$beta_pred[,2],pred$Upper[,2],pred$Lower[,2],beta0.2)
  # beta3
  plot.SCB(X0[,3],pred$beta_pred[,3],pred$Upper[,3],pred$Lower[,3],beta0.3)
  
  return(res)
}

# Simulation(2)
result=mclapply(1:2,Simulation,mc.cores=1)
result=do.call(rbind,result)
colMeans(result)
