######################################################################
# Project: Estimation and Inference for Generalized Geoadditive Models
# Author: Shan Yu, Guannan Wang, Li Wang, Chenhui Liu and Lijian Yang
# Date: November 2018
######################################################################
# Function: ggam.pred
# the function to fit the ggam
# input:
# (1) object: the fitted model
# (2) X: the covariate matrix. Each column represents one covariate.
# (3) S: the location matrix. Each row indicates the loacation of one point.
# (4) Tr: the triangulation matrix. Each row indicates one triangle.
# (5) V: the vertices matrix of the triangulation.
# (6) testlevel: confidence level to calculate the SCBs.
# Output:
# (1) alpha_pred: the predicted alpha function
# (2) alpha.ind: arranged order of S
# (3) beta_pred: the predicted beta function
# (4) Upper: upper bound
# (5) Lower: lower bound

ggam.pred=function(object.S,X,S,V,Tr,backfitting=TRUE,testlevel){

  # generate basis function for S using bivariate spline
  B0=basis(V,Tr,object.S$d,object.S$r,S)
  H=B0$H # smoothness and qr decomposition
  Q2=qrH(H)
  P=B0$K # energy matrix P
  P=t(Q2)%*%P%*%Q2
  Bi=B0$Bi # bivariate spline basis matrix
  BQ2=as.matrix(Bi%*%Q2)
  mm=BQ2%*%object.S$theta_hat
  Ind=B0$Ind
  X.p=dim(X)[2]

  if(backfitting==FALSE){
    # univariate spline
    XX=NULL
    p=NULL
    for(alpha in 1:X.p){
      Xalpha=Basis_generator(X[,alpha],N=N,q=rho+1,knots=knots[,alpha])$Bx
      XX=cbind(XX,Xalpha)
      p=c(p,dim(Xalpha)[2])
    }
    XX=as.matrix(XX)

    # spline estimator
    beta=object.S$beta_hat
    p=c(0,cumsum(p))
    mhat=NULL
    for(alpha in 1:X.p){
      m=XX[,(p[alpha]+1):(p[alpha+1])]%*%beta[(p[alpha]+1):(p[alpha+1])]
      mhat=cbind(mhat,m)
    }
    returnResults=list(alpha_pred=mm,beta_pred=mhat)
  }

  if(backfitting==TRUE){
    mhat=Upper=Lower=NULL
    H_opt=object.S$opt.bandwidth
    y=object.S$y
    n=length(y)

    for(alpha in 1:X.p){
      h_opt=H_opt[alpha]
      x0=X[,alpha]
      initial=runif(length(x0))-0.5

      m1_sbk=SBK_locp(y,object.S$X,initial,alpha,x0,object.S$mhat,h_opt,object.S$theta)

      bandwidth.option=object.S$bandwidth.option

      if(bandwidth.option==1) {
        aa=key_functionsSBLL(y,object.S$X,alpha,object.S$mhat,object.S$sigma_2,x0,
                             object.S$theta)
      } else if (bandwidth.option==2) {
        aa=key_functionsSBLL2(y,object.S$X,alpha,object.S$mhat,object.S$sigma_2,x0,
                              object.S$theta)
      } else {
        aa=key_functionsSBLL3(y,object.S$X,alpha,object.S$mhat,object.S$sigma_2,x0,
                              object.S$theta)
      }

      a_h=sqrt(-2*log(h_opt))
      q_testlevel=a_h+(log(3^0.5)-log(2*pi)-log(-log(1-testlevel)/2))/a_h
      upper=q_testlevel/sqrt(n*h_opt)*sqrt(1/aa$esigma/aa$df_fixed*5/7)+m1_sbk
      lower=-q_testlevel/sqrt(n*h_opt)*sqrt(1/aa$esigma/aa$df_fixed*5/7)+m1_sbk

      mhat=cbind(mhat,m1_sbk)
      Upper=cbind(Upper,upper)
      Lower=cbind(Lower,lower)
    }
    returnResults=list(alpha_pred=mm,alpha.ind=Ind,beta_pred=mhat,Upper=Upper,Lower=Lower)
  }
}
