######################################################################
# Project: Estimation and Inference for Generalized Geoadditive Models
# Author: Shan Yu, Guannan Wang, Li Wang, Chenhui Liu and Lijian Yang
# Date: November 2018
######################################################################
# Function: key_functions
# estimating key functions for bandwidth and confidence band
#
# Input Arguments:
#	(1) y: response variable.
# (2) X: Covariates.
# (3) alpha: indicating which function to estimate
#	(4) N: Number of interior knots in cubic splines
#	(5) xfixed: evaluated at xfixed.
#
# Output Arguments:
# (1) D: D(xfixed)
# (2) v2: v2(xfixed)
# (3) bias: bias(xfixed)
#######################################################################
library('splines')
key_functions=function(y,X,alpha,mhat,nn,xfixed,m=4,theta=0){
  df=approxfun(density(X[,alpha],kernel='biweight'))
  #step 0.1: basis generator
  knots=Basis_generator(X[,alpha],N=nn)$knots
  knots0=knots[2:(length(knots)-1)]
  XX=bs(X[,alpha],knots=knots0,degree=m-1,intercept=TRUE)
  #step 0.2: density estimation

  #### estimating ####
  #step 1.1
  a_coef=function(a){
    m=rowSums(mhat[,-alpha])+XX%*%a
    -sum(y*m-exp(m))
  }
  a=rep(0,length(knots0)+m)
  ahat=optim(a,a_coef)$par

  #step 1.2
  a2_coef=function(a){
    m=exp(rowSums(mhat))-XX%*%a
    sum(m^2)
  }
  a=rep(0,length(knots0)+m)
  ahat2=optim(a,a2_coef)$par


  #### evaluation ####
  XX=bs(xfixed,knots=knots0,degree=m-1,intercept=TRUE)
  XX1=bs(xfixed,knots=knots0,degree=m-2,intercept=TRUE)
  XX2=bs(xfixed,knots=knots0,degree=m-3,intercept=TRUE)
  D1=dev_coeff(m,knots0,1)
  D2=dev_coeff(m,knots0,2)

  m_dev1=XX1%*%(D1%*%ahat)
  m_dev2=XX2%*%(D2%*%ahat)
  eb_dev2=XX%*%ahat2
  eb_dev2_dev1=XX1%*%(D1%*%ahat2)
  eb_dev3=XX%*%ahat2


  xalpha=data.matrix(X[,alpha])
  n=nrow(X)
  m=length(xfixed)
  # calculate distance matrix
  Sample=matrix(rep(xalpha,each=m),ncol=m,byrow=TRUE)
  fix=matrix(rep(xfixed,each=n),nrow=n)
  ualpha=Sample-fix
  # kernel: biweight
  halpha=0.2
  kalpha=15/16*((1-(ualpha/halpha)^2+abs(1-(ualpha/halpha)^2))/2)^2/halpha

  sigma=(y-exp(rowSums(mhat)))^2
  #esigma=XX%*%solve(t(XX)%*%XX,t(XX)%*%sigma)
  esigma=t(sigma)%*%kalpha/colSums(kalpha)
  esigma=as.vector(esigma)


  if(theta==0) b_dev2=exp(rowSums(mhat)) else{
    b_dev2=exp(rowSums(mhat))
    b_dev2=exp(rowSums(mhat))+exp(rowSums(mhat))^2/theta
  }

  eb_dev2=(b_dev2%*%kalpha)/colSums(kalpha)

  if(theta==0) b_dev3=b_dev2 else{
    b_dev3=(2*exp(rowSums(mhat))/theta+1)*b_dev2
  }
  eb_dev2=as.vector(eb_dev2)
  eb_dev3=as.vector(eb_dev3)

  #### D part ####
  D=df(xfixed)*eb_dev2
  #### v^2 part ####
  v2=df(xfixed)*esigma*5/7
  #### bias part ####
  bias1=m_dev2*df(xfixed)*eb_dev2
  bias2=m_dev1*df(xfixed)*eb_dev2_dev1
  bias3=m_dev1^2*df(xfixed)*eb_dev3
  bias=1/7*(bias1+bias2-bias3)

  trun=function(x) max(x,0)
  D=abs(D)
  v2=sapply(v2,trun)

  list(D=D,v2=v2,bias=bias)
}


